import{S as A,i as M,s as E,E as f,Z as h,_,m,h as o,n as R,$ as g,b as z,a0 as p,L as B,a1 as c,N as v,Q as b,q as C,r as k,u as q}from"./index.d15f259d.js";function w(a){let e,s;return{c(){e=h("title"),s=C(a[1])},l(t){e=_(t,"title",{});var r=m(e);s=k(r,a[1]),r.forEach(o)},m(t,r){z(t,e,r),p(e,s)},p(t,r){r&2&&q(s,t[1])},d(t){t&&o(e)}}}function x(a){let e,s,t=a[1]&&w(a),r=[{xmlns:"http://www.w3.org/2000/svg"},{viewBox:"0 0 32 32"},{fill:"currentColor"},{preserveAspectRatio:"xMidYMid meet"},{width:a[0]},{height:a[0]},a[2],a[3]],n={};for(let i=0;i<r.length;i+=1)n=f(n,r[i]);return{c(){e=h("svg"),t&&t.c(),s=h("path"),this.h()},l(i){e=_(i,"svg",{xmlns:!0,viewBox:!0,fill:!0,preserveAspectRatio:!0,width:!0,height:!0});var l=m(e);t&&t.l(l),s=_(l,"path",{d:!0}),m(s).forEach(o),l.forEach(o),this.h()},h(){R(s,"d","M10 6L10 8 22.59 8 6 24.59 7.41 26 24 9.41 24 22 26 22 26 6 10 6z"),g(e,n)},m(i,l){z(i,e,l),t&&t.m(e,null),p(e,s)},p(i,[l]){i[1]?t?t.p(i,l):(t=w(i),t.c(),t.m(e,s)):t&&(t.d(1),t=null),g(e,n=B(r,[{xmlns:"http://www.w3.org/2000/svg"},{viewBox:"0 0 32 32"},{fill:"currentColor"},{preserveAspectRatio:"xMidYMid meet"},l&1&&{width:i[0]},l&1&&{height:i[0]},l&4&&i[2],l&8&&i[3]]))},i:c,o:c,d(i){i&&o(e),t&&t.d()}}}function L(a,e,s){let t,r;const n=["size","title"];let i=v(e,n),{size:l=16}=e,{title:d=void 0}=e;return a.$$set=u=>{s(5,e=f(f({},e),b(u))),s(3,i=v(e,n)),"size"in u&&s(0,l=u.size),"title"in u&&s(1,d=u.title)},a.$$.update=()=>{s(4,t=e["aria-label"]||e["aria-labelledby"]||d),s(2,r={"aria-hidden":t?void 0:!0,role:t?"img":void 0,focusable:Number(e.tabindex)===0?!0:void 0})},e=b(e),[l,d,r,i,t]}class S extends A{constructor(e){super(),M(this,e,L,x,E,{size:0,title:1})}}export{S as A};
