.new-form .bx--form-item{margin-bottom:1.5rem}.bx--fieldset{border-radius:.25rem;border-width:1px;--tw-border-opacity:1;border-color:rgb(209 213 219 / var(--tw-border-opacity));padding:1rem}legend.bx--label{padding-left:.5rem;padding-right:.5rem;font-size:1.125rem;line-height:1.75rem}
